#include "Utility.h"
#include "Global.h"
#include "Index.h"
#include "FiberGraph.h"
#include "Block.h"

#include <queue>

struct graph_partition {
	std::vector<coordinate>* coords = NULL;
	FiberGraph* graph = NULL;

	graph_partition() {
		//Do nothing
	}

	graph_partition(FiberGraph* parent, intVector* candidates, bool deleteParent) {
		graph = createPartitonGraph(parent->coordinates, candidates);
//		LOG_DB("creation coordinates = " + toString(graph->coordinates))
		coords = graph->coordinates;
		if (deleteParent)
			delete parent;
	}

	void deleteGraphs() {
		delete graph;
		delete coords;
	}

	~graph_partition() {
		//delete graph;
		//delete coords;
	}
};

inline float dprod(int min, int max) {
	float prod = 1.0f;
	for (float i = min; i < max; i = i + 1.0f) {
		prod *= (1.0f + (1.0f / i));
	}
	return prod;
}

float getSubBlockMaxDense(intSet* X, intSet* Y, intSet* Z) {
	int tp = global::BF->checkDensityTp(*X, *Y, *Z);
	float denseBase = ((float) tp)
			/ ((float) (X->size() * Y->size() * Z->size()));
	float maxDense = denseBase * dprod(global::BLOCK_MINSIZE, X->size())
			* dprod(global::BLOCK_MINSIZE, Y->size())
			* dprod(global::BLOCK_MINSIZE, Z->size());
	return maxDense;
}

intVector* selectBlockCandidatesFast(FiberGraph* graph) {
	intVector* result = new intVector();
	for (int i = 0; i < graph->numNodes; i++) {
		const intSet* candNeighbours = graph->getNeighbours(i);
		if (candNeighbours->size() >= global::BLOCK_MINSIZE) {
			result->push_back(i);
		}
	}
	return result;
}

intVector* selectByGraphNeighboursLowerFast(FiberGraph* graph,
		intVector* candidates, int minNeighbours) {
	intVector* result = new intVector();
	for (int cand : *candidates) {
		const intSet* candNeighbours = graph->getNeighbours(cand);
		if (candNeighbours->size() >= minNeighbours) {
			result->push_back(cand);
		}
	}
	return result;
}

intVector* selectBlockCandidatesFast(FiberGraph* graph, intVector* candidates) {
	return selectByGraphNeighboursLowerFast(graph, candidates,
			global::BLOCK_MINSIZE);
}

intVector* expand(FiberGraph* graph, intVector* candidates, BloomFilter* BF) {
	intVector* coreIds = new intVector();
	intSet Xcore;
	intSet Ycore;
	intSet Zcore;
	intSet removedNodes;
	std::queue<int> expansion;
	for (int nid : *candidates) {
		coreIds->push_back(nid);
		Xcore.insert((*(graph->coordinates))[nid].x);
		Ycore.insert((*(graph->coordinates))[nid].y);
		Zcore.insert((*(graph->coordinates))[nid].z);
		removedNodes.insert(nid);
	}

//	//Logging / Debug Info
//	int xOldSize = Xcore.size();
//	int yOldSize = Ycore.size();
//	int zOldSize = Zcore.size();
//	int numAdded = 0;

	for (int pivot : *candidates) {
		for (int nid : *(graph->getNeighbours(pivot))) {
			if (!SET_CONTAINS(nid, removedNodes)) {
				expansion.push(nid);
			}
		}
	}
	while (expansion.size() != 0) {
		intSet* Xcand;
		intSet* Ycand;
		intSet* Zcand;
		int newNode = expansion.front();
		expansion.pop();
		int x = (*(graph->coordinates))[newNode].x;
		int y = (*(graph->coordinates))[newNode].y;
		int z = (*(graph->coordinates))[newNode].z;
		if (SET_CONTAINS(newNode, removedNodes)) {
			continue;
		}
		removedNodes.insert(newNode);
		bool inXcore = SET_CONTAINS(x, Xcore);
		bool inYcore = SET_CONTAINS(y, Ycore);
		bool inZcore = SET_CONTAINS(z, Zcore);
		if (inXcore && inYcore && inZcore) {
			coreIds->push_back(newNode);
//			numAdded++;
			removedNodes.insert(newNode);
			for (int nid : *(graph->getNeighbours(newNode))) {
				if (!(SET_CONTAINS(newNode, removedNodes)))
					expansion.push(newNode);
			}
			continue;
		}
		if (!inXcore) {
			Xcand = new intSet(); //Memory Leak, but annoying to fix
			Xcand->insert(x);
		} else {
			Xcand = &Xcore;
		}
		if (!inYcore) {
			Ycand = new intSet(); //Memory Leak, but annoying to fix
			Ycand->insert(y);
		} else {
			Ycand = &Ycore;
		}
		if (!inZcore) {
			Zcand = new intSet(); //Memory Leak, but annoying to fix
			Zcand->insert(z);
		} else {
			Zcand = &Zcore;
		}
		if (BF->checkDensity(*Xcand, *Ycand, *Zcand, global::BLOCK_THRESHOLD)) {
			coreIds->push_back(newNode);
//			numAdded++;
			removedNodes.insert(newNode);
			Xcore.insert(x);
			Ycore.insert(y);
			Zcore.insert(z);
			for (int nid : *(graph->getNeighbours(newNode))) {
				if (!(SET_CONTAINS(newNode, removedNodes)))
					expansion.push(newNode);
			}
			if (!(BF->checkDensity(Xcore, Ycore, Zcore, global::BLOCK_THRESHOLD))) {
				LOG_WAR(
						"not dense anymore! - This could be an error in the code...")
//				if (BF->checkDensity(*Xcand, *Ycand, *Zcand, global::BLOCK_THRESHOLD)) {
////					LOG_DB("core = " + toString(Xcore) + " , " + toString(Ycore) + " , " + toString(Zcore))
////					LOG_DB("new = "+ toString(Xcand) + " , " + toString(Ycand) + " , " + toString(Zcand))
//				}
			}
		} else {
			removedNodes.insert(newNode);
		}

		//Tidy Up
//		if (inXcore)
//			delete Xcand;
//		if (inYcore)
//			delete Ycand;
//		if (inZcore)
//			delete Zcand;
	}
//	if (numAdded > 0)
//		LOG_DB(
//				" expandet from " + toString(xOldSize) + " -by- "
//						+ toString(yOldSize) + " -by- " + toString(zOldSize)
//						+ " to " + toString(Xcore.size()) + " -by- "
//						+ toString(Ycore.size()) + " -by- "
//						+ toString(Zcore.size()) + " while adding "
//						+ toString(numAdded) + " entries.")
	return coreIds;
}

void loadFromFile() {

	std::ifstream infile(global::INPUT_FILE.c_str());
	LOG_INFO("Reading the input file " + global::INPUT_FILE + "...");
	if (!infile.good()) {
		LOG_ERR("input file is no good!");
		throw std::runtime_error("input file could not be read!");
	}

	std::string line;
	int lineNumber = 0;
	int coordID = 0;
	int x_tmp, y_tmp, z_tmp;

	while (std::getline(infile, line)) {
		std::istringstream iss(line);
		if (lineNumber == 0) {
			if (!(iss >> global::MAX_X >> global::MAX_Y >> global::MAX_Z)) {
				throw std::runtime_error(
						"Could not read max parameters from input file!");
			}
			LOG_DB(
					"MAX_X = " + toString(global::MAX_X) + " | MAX_Y = "
							+ toString(global::MAX_X) + " | MAX_Z = "
							+ toString(global::MAX_Z));
		} else if (lineNumber == 1) {
			if (!(iss >> global::SIZE_TENSOR)) {
				throw std::runtime_error(
						"Could not size parameter from input file!");
			}
			global::coordinates.resize(global::SIZE_TENSOR);
			LOG_DB("SIZE_TENSOR = " + toString(global::SIZE_TENSOR))
		} else if (lineNumber > 1) {
			if (!(iss >> x_tmp >> y_tmp >> z_tmp)) {
				throw std::runtime_error(
						"Could not read coordinate from input file in line "
								+ toString(lineNumber) + "!");
			}
			global::coordinates[coordID++].update(x_tmp, y_tmp, z_tmp);
		}
		lineNumber++;
	}
	LOG_INFO("Reading of input file complete!");
}

void runSplitNExpand() {
	//Start Preprocessing!
	LOG_INFO("Start Preprocessing!");
#pragma omp parallel sections
	{
#pragma omp section
		{
			/* Bloom filter for original data */
			global::BF = new BloomFilter(global::BLOOM_CAPACITY,
					global::BLOOM_ERROR_RATE);
			for (coordinate coord : global::coordinates) {
				int x = coord.x;
				int y = coord.y;
				int z = coord.z;
				global::BF->add(x, y, z);
			}
		}
#pragma omp section
		{
			/* Bloom filter for covered data */
			global::BFmask = new BloomFilter(global::BLOOM_CAPACITY,
					global::BLOOM_ERROR_RATE);
			for (coordinate coord : global::coordinates) {
				int x = coord.x;
				int y = coord.y;
				int z = coord.z;
				global::BFmask->add(x, y, z);
			}
		}
#pragma omp section
		{
			/* Global Fiber Graph */
			global::Graph = new FiberGraph(&global::coordinates);
		}
	}
	BlockIndex* blocks = new BlockIndex(0);
	LOG_INFO("Preprocessing complete!")

	// Partitioning based on Connected Components
	LOG_INFO("Starting Partitioning...")
	int numOmitted = 0;
	std::vector<bool> coveredIds(global::SIZE_TENSOR);
	for (int i = 0; i < global::SIZE_TENSOR; i++) {
		coveredIds[i] = false;
	}
	intVector* candidates = rangeVec(global::SIZE_TENSOR);
	std::random_shuffle(candidates->begin(), candidates->end());
	std::queue<graph_partition> partitions;
	int lastPrinted = 0;
	int onePerNodes = 0.01f * ((float) global::SIZE_TENSOR);
	int numCovered = 0;
	for (int startNode = 0; startNode < global::SIZE_TENSOR; startNode++) {
		if (lastPrinted + onePerNodes < numCovered) {
			LOG_INFO(
					"progress "
							+ toString(
									100 * ((float ) numCovered)
											/ ((float )global::SIZE_TENSOR))
							+ "%...")
			LOG_INFO(
					"covered: " + toString(numCovered) + " ,numPartitions: "
							+ toString(partitions.size()))
			lastPrinted = numCovered;
		}
		if (coveredIds[startNode])
			continue;
		// Find all nodes of the current partition in a breath first search
		intVector* curPartition = new intVector();
		std::queue<int> openNodes;
		openNodes.push(startNode);
		coveredIds[startNode] = true;
		curPartition->push_back(startNode);
		numCovered++;
		while (openNodes.size() > 0) {
			int curNode = openNodes.front();
			openNodes.pop();
			const intSet* neighbours = global::Graph->getNeighbours(curNode);
			for (int neighId : *neighbours) {
				if (!coveredIds[neighId]) {
					openNodes.push(neighId);
					curPartition->push_back(neighId);
					coveredIds[neighId] = true;
					numCovered++;
				}
			}
			// Print Progress
			if (lastPrinted + onePerNodes < numCovered) {
				LOG_INFO(
						"progress "
								+ toString(
										100 * ((float ) numCovered)
												/ ((float )global::SIZE_TENSOR))
								+ "%...")
				lastPrinted = numCovered;
			}
		}
		if (curPartition->size() < global::BLOCK_MINVOL) {
			numOmitted += curPartition->size();
			LOG_TR("Omitted because of Volume")
			delete curPartition;
			continue;
		}
		// Check Size constraint
		if (curPartition->size()
				< global::BLOCK_MINSIZE * global::BLOCK_MINSIZE
						* global::BLOCK_MINSIZE) {
			numOmitted += curPartition->size();
			LOG_TR("Omitted because of Size")
			delete curPartition;
			continue;
		}
		// Don't check for really large partitions
		if (curPartition->size() < 0.1 * global::SIZE_TENSOR) {
			// Check the maxDensity constraint
			intSet Xcand;
			intSet Ycand;
			intSet Zcand;
			for (int coordId : *curPartition) {
				coordinate coord = (global::coordinates)[coordId];
				Xcand.insert(coord.x);
				Ycand.insert(coord.y);
				Zcand.insert(coord.z);
			}
			float maxDense = getSubBlockMaxDense(&Xcand, &Ycand, &Zcand);
			if (maxDense < global::BLOCK_THRESHOLD) {
				numOmitted += curPartition->size();
				LOG_TR("Omitted because of Max Density")
				delete curPartition;
				continue;
			}
		}
		// Create Local Coordinate Vector
		std::vector<coordinate>* coordinates = new std::vector<coordinate>(
				curPartition->size());
		int pos = 0;
		for (int coordId : *curPartition) {
			(*coordinates)[pos] = global::coordinates[coordId];
			pos++;
		}
		// Check for potential obvious blocks in partition, don't explore partitions with blocks further.
		intVector* candidates = selectBlockCandidatesFast(global::Graph,
				curPartition);
		bool foundBlock = false;
		if (curPartition->size() < 0.1 * global::SIZE_TENSOR)
			foundBlock = blocks->update(&global::coordinates, candidates,
					global::BFmask);

		if (!foundBlock) {
			graph_partition toAdd(global::Graph, curPartition, false);
			partitions.push(toAdd);
		} else {
			numOmitted += curPartition->size();
			delete curPartition;
		}
	}
	delete candidates;
	LOG_INFO(
			"Partitioning Data into " + toString(partitions.size())
					+ " parts complete. " + toString(numOmitted)
					+ " nodes where omitted! ")
	// Splitting Phase

#pragma omp parallel shared(partitions)
	while (partitions.size() > 0) {
		graph_partition curPartition;
		bool end = false;
#pragma omp critical
		{
			if (partitions.size() != 0) {
				curPartition = partitions.front();
				partitions.pop();
			} else {
				end = true;
			}
			LOG_INFO(
					"Progress: " + toString(partitions.size())
							+ " partitions to go!");
		}
		if (end) {
			continue;
		}
		if (curPartition.coords == NULL)
			continue;

		// Test for Volume Constraint
		if (curPartition.graph->numNodes < global::BLOCK_MINVOL) {
			curPartition.deleteGraphs();
			continue;
		}

		// Test for Size Constraint
		if (curPartition.graph->numNodes
				< global::BLOCK_MINSIZE * global::BLOCK_MINSIZE
						* global::BLOCK_MINSIZE) {
			curPartition.deleteGraphs();
			continue;
		}

		// Test Max Density of Block
		intSet Xcand;
		intSet Ycand;
		intSet Zcand;
		for (auto coord : *curPartition.coords) {
			Xcand.insert(coord.x);
			Ycand.insert(coord.y);
			Zcand.insert(coord.z);
		}
		float maxDense = getSubBlockMaxDense(&Xcand, &Ycand, &Zcand);
		if (maxDense < global::BLOCK_THRESHOLD) {
			curPartition.deleteGraphs();
			continue;
		}

		// Test For Block Candidate
		intVector* blockCandidates = selectBlockCandidatesFast(
				curPartition.graph);
		if (blocks->update(curPartition.coords, blockCandidates,
				global::BFmask)) {
			intVector* expBlockDandidates = expand(curPartition.graph,
					blockCandidates, global::BFmask);
			blocks->update(curPartition.coords, blockCandidates,
					global::BFmask);
			delete blockCandidates;
			curPartition.deleteGraphs();
			continue;
		}

		// Calculate Split with pseudo diameter
		auto cut = curPartition.graph->split();

		// If pseudo diameter is small enough expand, elso add split to partitions
		if (cut.diameter < global::SPLIT_MIN_DIAM) {
			std::vector<intVector*> nBlocks(global::EXPAND_NUM_SEEDS);

			// Expand #EXPAND_NUM_SEEDS random entries
			for (int i = 0; i < global::EXPAND_NUM_SEEDS; i++) {
				int seed = rand() % curPartition.graph->numNodes;
				intVector cands;
				cands.push_back(seed);
				nBlocks[i] = expand(curPartition.graph, &cands, global::BFmask);
			}
#pragma omp critical
			{
				// Recheck expanded blocks
				for (int i = 0; i < global::EXPAND_NUM_SEEDS; i++) {
					blocks->update(curPartition.graph->coordinates, nBlocks[i],
							global::BFmask);
				}
			}
		} else {

			// Add splits to partitions
			LOG_TR("SPLITTING size = " + partitions.size())

			// FIXME: MASSIVE MEMORY LEAKS here, but fix is slower...
			graph_partition firstPart(curPartition.graph, cut.first, false);
			graph_partition secPart(curPartition.graph, cut.second, true);
#pragma omp critical
			{
				partitions.push(firstPart);
				partitions.push(secPart);
			}
		}
	}
	LOG_INFO("finished Split'n'Expand")

	//Writing Blocks to File
	blocks->toFile(global::OUTPUT_FILE);

	//Tidy Up a bit
	delete blocks;
}

int main(int argc, char* argv[]) {
	//Scanning Arguments
	for (int i = 0; i < argc; i++) {
		std::string token = std::string(argv[i]);
		LOG_DB(token)
		if (token.compare("-infile") == 0 || token.compare("-file") == 0
				|| token.compare("-input") == 0 || token.compare("-f") == 0) {
			global::INPUT_FILE = std::string(argv[++i]);
		} else if (token.compare("-outfile") == 0 || token.compare("-out") == 0
				|| token.compare("-o") == 0) {
			global::OUTPUT_FILE = std::string(argv[++i]);
		} else if (token.compare("-block_dense") == 0
				|| token.compare("-density") == 0
				|| token.compare("-thresh") == 0 || token.compare("-d") == 0) {
			global::BLOCK_THRESHOLD = std::atof(argv[++i]);
		} else if (token.compare("-block_size") == 0
				|| token.compare("-block_min_size") == 0
				|| token.compare("-min_size") == 0
				|| token.compare("-minSize") == 0) {
			global::BLOCK_MINSIZE = std::atoi(argv[++i]);
		} else if (token.compare("-block_vol") == 0
				|| token.compare("-block_min_vol") == 0
				|| token.compare("-min_vol") == 0
				|| token.compare("-minVol") == 0) {
			global::BLOCK_MINVOL = std::atoi(argv[++i]);
		} else if (token.compare("-exp_seeds") == 0
				|| token.compare("-exp_num_seeds") == 0) {
			global::EXPAND_NUM_SEEDS = std::atoi(argv[++i]);
		} else if (token.compare("-min_diam") == 0
				|| token.compare("-split_min_diam") == 0) {
			global::SPLIT_MIN_DIAM = std::atoi(argv[++i]);
		}
	}
	//Check Argument Sanity
	if (global::INPUT_FILE.compare(INPUT_FILE_DEFAULT) == 0) {
		throw std::runtime_error(
				"Input File has to be Specified! (use either -infile, -file, -input, -f)");
	}
	if (global::BLOCK_THRESHOLD > 1 || global::BLOCK_THRESHOLD < 0) {
		throw std::runtime_error("Block Threshold has to be  between 0 and 1!");
	}
	if (global::BLOCK_THRESHOLD < 0.5) {
		LOG_WAR(
				"Block threshold is less than 0.5! This may lead to suboptimal results!")
	}
	if (global::BLOCK_MINSIZE < 1) {
		throw std::runtime_error("Block minSize has to be at least 1!");
	}
	if (global::BLOCK_MINVOL < 1) {
		throw std::runtime_error("Block minVol should be at least 1!");
	}
	if (global::EXPAND_NUM_SEEDS < 0) {
		throw std::runtime_error("EXPAND num seeds should be larger equal 0!");
	}
	if (global::SPLIT_MIN_DIAM < 0) {
		throw std::runtime_error("SPLIT MIN DIAMETER should be larger equal 0!");
	}
	if (global::BLOCK_MINVOL
			< global::BLOCK_MINSIZE * global::BLOCK_MINSIZE
					* global::BLOCK_MINSIZE) {
		LOG_WAR("Block minVol is less than minSize^3. This doesn't make sense!")
	}
	//Print Final Arguments:
	std::cout << global::allToString() << std::endl;
	//Scan Tensor
	loadFromFile();
	//Run Algorithm
	runSplitNExpand();
	//Finish!
	return EXIT_SUCCESS;
}

